using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Networking;
using Newtonsoft.Json;
using TMPro;
using UnityEngine.EventSystems;

public class CreateAccountScript : MonoBehaviour
{
    [Header("Les différentes scènes")]
    [SerializeField] GameObject createAccountMenu;
    [SerializeField] GameObject loginMenu;
    [SerializeField] GameObject joinLobbyMenu;
    [SerializeField] GameObject settingsMenu;


    [Header("Les boutons de la scène")]
    [SerializeField] Button buttonCreate;
    [SerializeField] Button buttonReturn;
    [SerializeField] Button buttonSettings;


    [Header("Champs de saisi de données")]
    [SerializeField] GameObject userNameField;
    [SerializeField] GameObject passwordField;
    [SerializeField] GameObject confirmPasswordField;


    [Header("Message d'erreur")]
    public TMP_Text infoUserNameText;
    public TMP_Text infoPasswordText;
    public TMP_Text infoConfirmPasswordText;
    public TMP_Text infoServer;


    private string userName = "";
    private string password = "";
    private string confirmPassword = "";

    private static int minChar = 3;
    private static int maxChar = 20;

    private Regex regex;

    // Liste des champs de saisi de la scène
    private List<GameObject> fields;

    void Start()
    {
        // \s correspond à un whitespace 
        regex = new Regex(@"(\s)");

        // Défini les actions des boutons
        buttonCreate.onClick.AddListener(Click);
        buttonReturn.onClick.AddListener(delegate { GoFromMenuToMenu(createAccountMenu, loginMenu); });
        buttonSettings.onClick.AddListener(GoToSettingsMenu);

        // rempli la liste de champs de saisi
        fields = new List<GameObject> { userNameField, passwordField, confirmPasswordField };
    }

    public void Update()
    {
        // Lorsqu'on presse la touche tab, change le champ de saisi sélectionné
        if (Input.GetKeyDown(KeyCode.Tab))
        {
            EventSystem.current.SetSelectedGameObject(fields[(fields.IndexOf(EventSystem.current.currentSelectedGameObject) + 1) % fields.Count]);
        }

        // Lorsqu'on presse la touche entrée, simule le bouton
        if (Input.GetKeyDown(KeyCode.Return))
        {
            buttonCreate.onClick?.Invoke();
        }

        // Lorsqu'on presse la touche echap, retourne à l'accueil
        if (Input.GetKeyDown(KeyCode.Escape))
        {
            buttonReturn.onClick?.Invoke();
        }
    }

    // Déclanché dès que l'objet est activé (= SetActive(true))
    void OnEnable()
    {
        infoUserNameText.gameObject.SetActive(false);
        userNameField.GetComponent<TMP_InputField>().text = "";
        userName = "";

        infoPasswordText.gameObject.SetActive(false);
        passwordField.GetComponent<TMP_InputField>().text = "";
        password = "";

        infoConfirmPasswordText.gameObject.SetActive(false);
        confirmPasswordField.GetComponent<TMP_InputField>().text = "";
        confirmPassword = "";

        infoServer.gameObject.SetActive(false);
        infoServer.text = "";
    }

    // Evènement au click du bouton de connexion de compte
    public void Click()
    {
        // Désactive le message d'information du serveur
        infoServer.gameObject.SetActive(false);

        if (CheckUserName() && CheckPassword() && CheckConfirmPassword())
            StartCoroutine(GetText());
    }

    // Check si l'username est conforme ou non
    public bool CheckUserName()
    {
        infoUserNameText.gameObject.SetActive(false);

        // Récupère l'information contenu dans le champ de saisi 
        var s = userNameField.GetComponent<TMP_InputField>().text;

        if (s == "" || regex.IsMatch(s) || s.Length < minChar || s.Length > maxChar)
        {
            infoUserNameText.gameObject.SetActive(true);
            userName = "";
            return false;
        }
        else
        {
            userName = s;
            return true;
        }
    }

    // Check si le mot de passe est conforme ou non
    public bool CheckPassword()
    {
        infoPasswordText.gameObject.SetActive(false);

        // Récupère l'information contenu dans le champ de saisi 
        var s = passwordField.GetComponent<TMP_InputField>().text;

        if (s == "" || regex.IsMatch(s) || s.Length < minChar || s.Length > maxChar)
        {
            infoPasswordText.gameObject.SetActive(true);
            password = "";
            return false;
        }
        else
        {
            password = s;
            return true;
        }
    }

    // Check si le deuxième mot de passe correspond ou non
    public bool CheckConfirmPassword()
    {
        infoConfirmPasswordText.gameObject.SetActive(false);

        // Récupère l'information contenu dans le champ de saisi 
        var s = confirmPasswordField.GetComponent<TMP_InputField>().text;

        if (password != s || password == "" || s == "")
        {
            infoConfirmPasswordText.gameObject.SetActive(true);
            confirmPassword = "";
            return false;
        }
        else
        {
            confirmPassword = s;
            return true;
        }
    }

    // Envoie les données renseignées à la BDD, et traite l'information la réponse
    IEnumerator GetText()
    {
        // Créé un formulaire d'envoie spécifique pour être compris par la BDD
        WWWForm form = new WWWForm();
        form.AddField("userName", userName);
        form.AddField("password", password);

        // On communique les informations
        using (UnityWebRequest webRequest = UnityWebRequest.Post("https://mai-projet-integrateur.u-strasbg.fr/vmProjetIntegrateurgrp6-0/create_account.php", form))
        {
            webRequest.certificateHandler = new CertificateOverride();
            yield return webRequest.SendWebRequest();

            // La connexion au serveur a echoue
            if (webRequest.result == UnityWebRequest.Result.ConnectionError || webRequest.result == UnityWebRequest.Result.ProtocolError)
            {
                Debug.Log(webRequest.error);
            }
            // La connexion au serveur a reussi
            else
            {
                if (webRequest.isDone)
                {
                    // Converti la réponse en JSON
                    var requestRes = JsonConvert.DeserializeObject<Dictionary<string, string>>(webRequest.downloadHandler.text);
                    Debug.Log(webRequest.downloadHandler.text);


                    // Les identifiants renseignes par l'utilisateur sont corrects
                    if (requestRes["isSuccessful"] == "1")
                    {
                        PlayerInfo.SetUserName(userName);
                        PlayerInfo.SetConnection(true);
                        GoFromMenuToMenu(createAccountMenu, joinLobbyMenu);
                    }


                    // Les identifiants renseignes par l'utilisateur sont erronnes
                    else if (requestRes["isSuccessful"] == "0")
                    {
                        // le login renseigne existe deja dans la BDD
                        if (requestRes["errorCode"] == "500")
                        {
                            infoServer.gameObject.SetActive(true);
                            infoServer.text = "Le login existe déjà !";
                        }

                        // une erreur est survenue
                        if (requestRes["errorCode"] == "600")
                        {
                            infoServer.gameObject.SetActive(true);
                            infoServer.text = "Une erreur est survenue, veuillez réessayer";
                        }
                    }
                }
            }
        }
    }

    private void GoFromMenuToMenu(GameObject menuA, GameObject menuB)
    {
        menuA.SetActive(false);
        menuB.SetActive(true);
    }

    private void GoToSettingsMenu()
    {
        settingsMenu.SetActive(true);
    }
}
