/*
JavaSript server for the game,
by Hugo Himber and Vincent Cardile

source code used as base for this project :
https://github.com/bradstondevcode/Creating-A-Multiplayer-Game-Server-Series/blob/master/Episode-4/server.js
*/


const argument = process.argv[2];

if (argument !== undefined && argument !== "debug")
{
	console.log("argument not recognized, program stopped");
	return;
}

// if debugMode is true, the server will print extra information
const debugMode = (argument == "debug") ? true : false;

const WebSocket = require("ws").Server;

const ws = new WebSocket({port: 10003, 'Access-Control-Allow-Origin': "*"});


////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////	DATA OBJECTS	////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

/*
client <--> server for lobby joining then client --> server in game:
JSON PlayerDataSimple

// individual data
public string roomCode { get; set; }
public string createLobby { get; set; }
public string inLobby { get; set; }
public string playerNumber{ get; set; }
public string isReady{ get; set; }
public string errorCode { get; set; }
public string errorMsg { get; set; }

// data to be shared with other players
public string login { get; set; }
public string currentScene { get; set; }
public string hitPoint { get; set; }
public string posX { get; set; }
public string posY { get; set; }
public string posZ { get; set; }
public string rotY { get; set; }
public string animWalk { get; set; }
public string animAttack { get; set; }
public string animSpeAtt { get; set; }
*/
var PlayerDataSimple = [];

/*
server -> clients only:
JSON PlayerDataOnline

// data to be shared with all players
public string mainPlayer { get; set; }
public string currentScene { get; set; }
public string roomCode { get; set; }
public string errorCode { get; set; }
public string errorMsg { get; set; }

public string[] login { get; set; }
public string[] status { get; set; }
public string[] hitPoint { get; set; }
public string[] posX { get; set; }
public string[] posY { get; set; }
public string[] posZ { get; set; }
public string[] rotY { get; set; }
public string[] animWalk { get; set; }
public string[] animAttack { get; set; }
public string[] animSpeAtt { get; set; }
*/
var PlayerDataOnline = [];

/*
variable containing the data of lobbies:
- storage is direct data from the lobby that's updated by the clients sending
- players is kind of a "shortcut" containing useful player only data to speedup the computations on the server
List LobbyData

// data storage for the server
public PlayerDataOnline[] storage { get; set; }
public string[] players (login, type, roomCode, socket) { get; set; }
*/
var LobbyData = {storage: [], players: []};


// useful to compare if the current client is a player
var playersType = ["Warrior", "Archer"];


////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////	    FUNCTIONS   	////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

/**
 * Function returning the index of an element in a table
 * @param table, array to search in
 * @param key, index of the element to search
 * @param element, ellment to search
 * @returns int, index of the element in the table, -1 if not found
 */
function findIndex(table, key, element)											
{
	for (let i = 0; i < table.length; i++)
	{
		if (table[i][key] == element)
			return i;
	}
	return -1;
}

/**
 * Function printing the server data for debug purposes
 */
function printServerData()
{
	LobbyData.storage.forEach((lobby, index) => 
	{
		let nbPlayers = 0;

		for (let i = 0; i < lobby.login.length; i++)
			if (playersType.indexOf(lobby.type[i]) != -1)
				nbPlayers++;

		console.log("CONTENU ACTUEL SUR LE SERVEUR : LOBBY", index + 1);

		console.log("------ Lobby ------");
		console.log("roomCode :", lobby["roomCode"]);
		console.log("nbPlayers :", nbPlayers);
		console.log("mainPlayer :", lobby["mainPlayer"]);
		console.log("currentScene :", lobby["currentScene"]);

		console.log("------ Players ------");
		for (let i = 0; i < lobby.login.length; i++)
			console.log("login :", lobby.login[i], ", type :", lobby.type[i], ", isReady :", lobby.status[i]);
	});
}

/**
 * Function removing a player from a lobby
 * @param socket, websocket of the player
 */
function removePlayer(socket, loginGiven)
{
	// this MUST BE in a try catch to prevent the server from crashing if something goes wrong
	try 
	{
		let indexInPlayers = findIndex(LobbyData.players, "socket", socket);

		if (indexInPlayers == -1) 
		{
			if (debugMode) console.log("the player was not in a lobby, nothing changed");
			return;
		}

		// if a login is given (quitting the lobby) we use it, otherwise we use the socket to find it (deconnection)
		let roomCode = LobbyData.players[indexInPlayers]["roomCode"];
		let login =	(loginGiven == null ? LobbyData.players[indexInPlayers]["login"] : loginGiven)

		if (!login.startsWith("enemy" + roomCode))
			LobbyData.players.splice(indexInPlayers, 1);

		let indexInStorage = findIndex(LobbyData.storage, "roomCode", roomCode);
		let indexInStorageLogin;

		// find the index of the person/enemy who leave the lobby
		for (let i = 0; i < LobbyData.storage[indexInStorage].login.length; i++)
		{
			if (LobbyData.storage[indexInStorage].login[i] == login)
			{
				indexInStorageLogin = i;
				break;
			}
		}

		if (login.startsWith("enemy" + roomCode))
		{
			LobbyData.storage[indexInStorage].hitPoint[indexInStorageLogin] = "0";

			LobbyData.players.forEach((player, index) => 
			{
				if (player["roomCode"] == roomCode)
					player.socket.send(JSON.stringify(LobbyData.storage[indexInStorage]));

				if (debugMode) console.log("PAQUET ENVOYE AU JOUEUR", index + 1, "\n", LobbyData.storage[indexInStorage]);
			});
		}

		// if (PlayerDataSimple["type"] == "Melee Minion")
		// 	console.log("enemy deleted !", roomCode, login, indexInStorageLogin)

		// deleting his data
		LobbyData.storage[indexInStorage].login.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].type.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].status.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].hitPoint.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].posX.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].posY.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].posZ.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].rotY.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].animMoveForward.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].animMoveBackward.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].animMoveRight.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].animMoveLeft.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].animAttack.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].animSpeAtt.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].animRoll.splice(indexInStorageLogin, 1);
		LobbyData.storage[indexInStorage].animShield.splice(indexInStorageLogin, 1);

		// enemies don't get a response (because the player will take it !)
		if (!login.startsWith("enemy" + roomCode))
		{
			let nbPlayers = 0;

			for (let i = 0; i < LobbyData.storage[indexInStorage].login.length; i++)
				if (playersType.indexOf(LobbyData.storage[indexInStorage].type[i]) != -1)
					nbPlayers++;

			// the lobby is deleted if no one use it anymore
			if (nbPlayers == 0)
			{
				LobbyData.storage.splice(indexInStorage, 1);

				if (debugMode) console.log("success : lobby deleted because everyone left")

				return;
			}

			// if it was the main player who leave, the second player will take his role
			if (LobbyData.storage[indexInStorage]["mainPlayer"] == login)
			{
				LobbyData.storage[indexInStorage]["mainPlayer"] = LobbyData.storage[indexInStorage].login[0];
			}

			LobbyData.storage[indexInStorage]["errorCode"] = "40";
			LobbyData.storage[indexInStorage]["errorMsg"] = "playerDisconnection";

			// we inform other players from the lobby that a player leave it
			LobbyData.players.forEach((player, index) => 
			{
				if (player["roomCode"] == roomCode)
					player.socket.send(JSON.stringify(LobbyData.storage[indexInStorage]));

				if (debugMode) console.log("PAQUET ENVOYE AU JOUEUR", index + 1, "\n", LobbyData.storage[indexInStorage]);
			});

			LobbyData.storage[indexInStorage]["errorCode"] = "0";
			LobbyData.storage[indexInStorage]["errorMsg"] = "";
		}
	} 
	catch (error) 
	{
		console.log("Error while deleting player or enemy :", error);
	}
}

/*
 * Function that generate a 6 characters roomCode who does not already exists
 */
function generateRoomCode()
{
	let roomCodeValid = false;
	let roomCode = "";

	while (!roomCodeValid)
	{
		roomCode = "";

		for (let i = 0; i < 6; i++)
			roomCode += String.fromCharCode(65 + Math.random() * 26)

		if (findIndex(LobbyData.storage, "roomCode", roomCode) == -1)
			roomCodeValid = true;
	}

	return roomCode;
}

////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////	WEBSOCKET CONNECTION	////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

ws.on('connection', function connection(client) 
{
	if (debugMode) console.log("\n\nnew connection !");  // sanity check

	////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////	 COMMUNICATION 	////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////

	client.on("message", (data) =>
	{
		// parse the data
		PlayerDataSimple = JSON.parse(data);

		if (debugMode) console.log("\n\nPAQUET RECU :\n", PlayerDataSimple);

		// if we receive a message with no login, we don't even consider the message
		if (PlayerDataSimple["login"] == null)
			return;

		////////////////////////////////////////////////////////////////////////////////////////
		////////////////////////////	LOBBY CREATION	  //////////////////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////

		/**
		 * If createLobby is true, create a new lobby in the LobbyData storage (LobbyData.storage)
		 * @var PlayerDataSimple ,working data
		 * @send (error) 11, if we already are in a lobby
		 */

		// shouldn't be happening, but just in case
		if (PlayerDataSimple["createLobby"] === "true" && PlayerDataSimple["inLobby"] === "true")
		{
			PlayerDataSimple["errorCode"] = "11";
			PlayerDataSimple["errorMsg"] = "cannotCreateLobbyInLobby";
			client.send(JSON.stringify(PlayerDataSimple));

			if (debugMode) console.log("error : cannotCreateLobbyInLobby");
			if (debugMode) console.log("PAQUET ENVOYE :\n", dataJson);

			return;
		}

		// we want to create a lobby
		if (PlayerDataSimple["createLobby"] === "true")
		{
			let roomCode = generateRoomCode();
			let PlayerDataOnline = {mainPlayer: PlayerDataSimple["login"], currentScene: PlayerDataSimple["currentScene"], roomCode: roomCode, errorCode: "0", errorMsg: "",
									login: [], type: [], status: [], hitPoint: [], posX: [], posY: [], posZ: [], rotY: [], animMoveForward: [], animMoveBackward: [],
									animMoveRight: [], animMoveLeft: [], animAttack: [], animSpeAtt: [], animRoll: [], animShield: []};
			LobbyData.storage.push(PlayerDataOnline);

			PlayerDataSimple["roomCode"] = roomCode;
			PlayerDataSimple["createLobby"] = "false";

			if (debugMode) console.log("success : roomCreated, roomCode " + roomCode);
		}

		////////////////////////////////////////////////////////////////////////////////////////
		////////////////////////////	LOBBY QUITTING	  //////////////////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////

		/**
		 * Handle lobby quitting process
		 * if inLobby is false, roomCode is null (that's the code we decided to use),
		 * the player will quit the lobby (equivalent to removing the player)
		 * @var PlayerDataSimple, working data
		 * @send (error) 24, to inform the player that we quitted successfully
		 * @send (error) 40 to the others, to tell them that someone disconnected
		 */
		
		// if we want to quit the lobby
		if (PlayerDataSimple["inLobby"] === "false" && PlayerDataSimple["roomCode"] == "")
		{
			if (playersType.indexOf(PlayerDataSimple["type"]) != -1)
			{
				let PlayerDataOnline = {errorCode: "24", errorMsg: "lobbyQuitting"}
				client.send(JSON.stringify(PlayerDataOnline));

				if (debugMode) console.log("PAQUET ENVOYE :\n", PlayerDataOnline);
			}

			removePlayer(client, PlayerDataSimple["login"]);
			
			if (debugMode) console.log("success : lobby quitted successfully");
			if (debugMode) printServerData();
			
			return;
		}

		////////////////////////////////////////////////////////////////////////////////////////
		////////////////////////////	LOBBY JOINING	  //////////////////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////

		/**
		 * Handle lobby joining process
		 * if inLobby is false, indexInStorage != -1 and there is less than 4 players in the lobby,
		 * add the player to LobbyData.storage et create a "shortcut" for this player in LobbyData.players
		 * @var indexInStorage, index of the lobby in the LobbyData.storage array
		 * @var PlayerDataSimple ,working data
		 * @send (error) 22, if the lobby doesn't exist
		 * @send (error) 23 ,if the lobby is full
		 */

		// get the index of the lobby in the lobbyData array, if -1, the lobby doesn't exist or we want to quit the lobby
		let indexInStorage = findIndex(LobbyData.storage, "roomCode", PlayerDataSimple["roomCode"]);

		// if we try to join a lobby that does not exists
		if (indexInStorage == -1)
		{
			PlayerDataSimple["errorCode"] = "22";
			PlayerDataSimple["errorMsg"] = "lobbyNotFound";
			client.send(JSON.stringify(PlayerDataSimple));

			if (debugMode) console.log("error : lobbyNotFound, roomCode " + PlayerDataSimple["roomCode"]);
			if (debugMode) console.log("PAQUET ENVOYE :\n", PlayerDataSimple);

			return;
		}
		
		// if we want to join the lobby that exists
		if (PlayerDataSimple["inLobby"] === "false")
		{
			let nbPlayers = 0;

			for (let i = 0; i < LobbyData.storage[indexInStorage].login.length; i++)
				if (playersType.indexOf(LobbyData.storage[indexInStorage].type[i]) != -1)
					nbPlayers++;

			// only players should have join restriction
			if (playersType.indexOf(PlayerDataSimple["type"]) != -1)
			{
				// NOPE if we already are in a lobby (it should never be the case)
				if (findIndex(LobbyData.players, "login", PlayerDataSimple["login"]) != -1)
				{
					// CHANGER ICI !!!!!!!!!!!!
					PlayerDataSimple["errorCode"] = "24";
					PlayerDataSimple["errorMsg"] = "alreadyInALobby";
					client.send(JSON.stringify(PlayerDataSimple));

					if (debugMode) console.log("error : alreadyInALobby");
					if (debugMode) console.log("PAQUET ENVOYE :\n", PlayerDataSimple);

					return;
				}

				// NOPE if the players are already in the game
				if (LobbyData.storage[indexInStorage]["currentScene"] != "Lobby3D")
				{
					PlayerDataSimple["errorCode"] = "25";
					PlayerDataSimple["errorMsg"] = "playersAlreadyInGame";
					client.send(JSON.stringify(PlayerDataSimple));

					if (debugMode) console.log("error : playersAlreadyInGame");
					if (debugMode) console.log("PAQUET ENVOYE :\n", PlayerDataSimple);

					return;
				}

				// NOPE if there is too many player
				if (nbPlayers >= 4)
				{
					PlayerDataSimple["errorCode"] = "23";
					PlayerDataSimple["errorMsg"] = "roomFull";
					client.send(JSON.stringify(PlayerDataSimple));

					if (debugMode) console.log("error : roomFull, roomCode " + PlayerDataSimple["roomCode"]);
					if (debugMode) console.log("PAQUET ENVOYE :\n", PlayerDataSimple);

					return;
				}
			}

			// we finally join the lobby
			LobbyData.storage[indexInStorage].login.push(PlayerDataSimple["login"]);
			LobbyData.storage[indexInStorage].type.push(PlayerDataSimple["type"]);
			LobbyData.storage[indexInStorage].status.push(PlayerDataSimple["isReady"]);
			LobbyData.storage[indexInStorage].hitPoint.push(PlayerDataSimple["hitPoint"]);
			LobbyData.storage[indexInStorage].posX.push(PlayerDataSimple["posX"]);
			LobbyData.storage[indexInStorage].posY.push(PlayerDataSimple["posY"]);
			LobbyData.storage[indexInStorage].posZ.push(PlayerDataSimple["posZ"]);
			LobbyData.storage[indexInStorage].rotY.push(PlayerDataSimple["rotY"]);
			LobbyData.storage[indexInStorage].animMoveForward.push(PlayerDataSimple["animMoveForward"]);
			LobbyData.storage[indexInStorage].animMoveBackward.push(PlayerDataSimple["animMoveBackward"]);
			LobbyData.storage[indexInStorage].animMoveRight.push(PlayerDataSimple["animMoveRight"]);
			LobbyData.storage[indexInStorage].animMoveLeft.push(PlayerDataSimple["animMoveLeft"]);
			LobbyData.storage[indexInStorage].animAttack.push(PlayerDataSimple["animAttack"]);
			LobbyData.storage[indexInStorage].animSpeAtt.push(PlayerDataSimple["animSpeAtt"]);
			LobbyData.storage[indexInStorage].animRoll.push(PlayerDataSimple["animRoll"]);
			LobbyData.storage[indexInStorage].animShield.push(PlayerDataSimple["animShield"]);

			// only players get an answer so they can change scene
			if (playersType.indexOf(PlayerDataSimple["type"]) != -1)
			{
				PlayerDataSimple["playerNumber"] = (nbPlayers + 1).toString();
				PlayerDataSimple["inLobby"] = "true";
				client.send(JSON.stringify(PlayerDataSimple));

				if (debugMode) console.log("success : roomJoined, roomCode " + PlayerDataSimple["roomCode"]);
				if (debugMode) console.log("PAQUET ENVOYE :\n", PlayerDataSimple);
				if (debugMode) printServerData();

				let player = {login: PlayerDataSimple["login"], type: PlayerDataSimple["type"], roomCode: PlayerDataSimple["roomCode"], socket : client};
				LobbyData.players.push(player);
			}

			return;
		}

		////////////////////////////////////////////////////////////////////////////////////////
		////////////////////////////	  DATA RELAY  	  //////////////////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////

		// data from the current is updated according to the data that we receive
		let playerIndex = LobbyData.storage[indexInStorage].login.indexOf(PlayerDataSimple["login"]);

		// only the main player can update the scene to show to clients
		if (LobbyData.storage[indexInStorage].mainPlayer == PlayerDataSimple["login"])
			LobbyData.storage[indexInStorage].currentScene = PlayerDataSimple["currentScene"];

		LobbyData.storage[indexInStorage].login[playerIndex] = PlayerDataSimple["login"];
		LobbyData.storage[indexInStorage].type[playerIndex] = PlayerDataSimple["type"];
		LobbyData.storage[indexInStorage].status[playerIndex] = PlayerDataSimple["isReady"];
		LobbyData.storage[indexInStorage].hitPoint[playerIndex] = PlayerDataSimple["hitPoint"];
		LobbyData.storage[indexInStorage].posX[playerIndex] = PlayerDataSimple["posX"];
		LobbyData.storage[indexInStorage].posY[playerIndex] = PlayerDataSimple["posY"];
		LobbyData.storage[indexInStorage].posZ[playerIndex] = PlayerDataSimple["posZ"];
		LobbyData.storage[indexInStorage].rotY[playerIndex] = PlayerDataSimple["rotY"];
		LobbyData.storage[indexInStorage].animMoveForward[playerIndex] = PlayerDataSimple["animMoveForward"];
		LobbyData.storage[indexInStorage].animMoveBackward[playerIndex] = PlayerDataSimple["animMoveBackward"];
		LobbyData.storage[indexInStorage].animMoveRight[playerIndex] = PlayerDataSimple["animMoveRight"];
		LobbyData.storage[indexInStorage].animMoveLeft[playerIndex] = PlayerDataSimple["animMoveLeft"];
		LobbyData.storage[indexInStorage].animAttack[playerIndex] = PlayerDataSimple["animAttack"];
		LobbyData.storage[indexInStorage].animSpeAtt[playerIndex] = PlayerDataSimple["animSpeAtt"];
		LobbyData.storage[indexInStorage].animRoll[playerIndex] = PlayerDataSimple["animRoll"];
		LobbyData.storage[indexInStorage].animShield[playerIndex] = PlayerDataSimple["animShield"];

		PlayerDataOnline = LobbyData.storage[indexInStorage];

		// if there is an animation to play or an event to share, we alert all person in the lobby so they can play the animation or react to event
		if (playersType.indexOf(PlayerDataSimple["type"] != -1) && (LobbyData.storage[indexInStorage]["errorCode"] != 0 ||
		PlayerDataSimple["animAttack"] == "true" || PlayerDataSimple["animSpeAtt"] == "true" || PlayerDataSimple["animRoll"] == "true"))
		{
			LobbyData.players.forEach(player =>
			{
				if (player.roomCode === PlayerDataSimple["roomCode"])
				{
					player.socket.send(JSON.stringify(PlayerDataOnline));

					if (debugMode) console.log("PAQUET ENVOYE AUX JOUEURS DU LOBBY :\n", PlayerDataOnline);
				}
			});

			LobbyData.storage[indexInStorage]["errorCode"] = "0";
			LobbyData.storage[indexInStorage]["errorMsg"] = "";

			for (let i = 0; i < LobbyData.storage[indexInStorage].login.length; i++)
			{
				LobbyData.storage[indexInStorage].animAttack[i] = "false";
				LobbyData.storage[indexInStorage].animSpeAtt[i] = "false";
				LobbyData.storage[indexInStorage].animRoll[i] = "false";
			}

			if (debugMode) printServerData();
		}
		else if (playersType.indexOf(PlayerDataSimple["type"]) != -1)
		{
			client.send(JSON.stringify(PlayerDataOnline));

			if (debugMode) console.log("PAQUET ENVOYE AU JOUEUR :\n", PlayerDataOnline);
			if (debugMode) printServerData();
		}
	});
	
	////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// 	DISCONNECTION 	////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////

	// dirty disconnection
	client.on("disconnect", () =>
	{
		if (debugMode) console.log("\n\ndirty disconnect detected !");
		if (debugMode) console.log("warning : playerDisconnection");
		
		removePlayer(client, null);

		if (debugMode) console.log("success : player deleted");
		if (debugMode) printServerData();
	});

	// clean disconnection
	client.on("close", () =>
	{
		if (debugMode) console.log("\n\nclean disconnect detected !");
		if (debugMode) console.log("warning : playerDisconnection");

		removePlayer(client, null);
		
		if (debugMode) console.log("success : player deleted");
		if (debugMode) printServerData();
	});
});

////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////	NOTHING TO SEE	////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
ws.on("listening", () =>
{
	console.log("Server is alive !");
});