/*
JavaSript server for the game,
by Hugo Himber and Vincent Cardile

source code used as base for this project :
https://github.com/bradstondevcode/Creating-A-Multiplayer-Game-Server-Series/blob/master/Episode-4/server.js
*/


const argument = process.argv[2];

if (argument !== undefined && argument !== "debug")
{
	console.log("argument not recognized, program stopped");
	return;
}

// if debugMode is true, the server will print extra information
const debugMode = (argument == "debug") ? true : false;

const fs = require("fs");
const https = require("https");
const WebSocket = require("ws");

/*
Commands to create SSL certificates :
openssl genrsa -out key.pem 2048
openssl req -new -key key.pem -out csr.pem
openssl x509 -req -days 3650 -in csr.pem -signkey key.pem -out cert.pem
*/

// loading SSL certificates
const privateKey = fs.readFileSync("credentials/key.pem", "utf8");
const certificate = fs.readFileSync("credentials/cert.pem", "utf8");
const credentials = {key: privateKey, cert: certificate};

// create the websocket server
const httpServer = https.createServer(credentials);
const wss = new WebSocket.Server({server: httpServer});


////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////	DATA OBJECTS	////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

/*
JSON dataJson
variable contening the data trensiting between the client and the server :
- string login,
- string roomCode,
- bool createLobby,
- bool inLobby,
- bool isReady,
- int errorCode,
- string errorMsg,
- int playerNumber,
- string[] playersInLobby,
- string[] statusInLobby,
- [data]
*/
var dataJson = [];

/*
Liste lobbyData
variable contening the data of the lobbies serverside :
- string roomCode : code of the lobby
- int nbPlayers : number of players in the lobby
- players[] (login, isReady, socket) : logins and sockets of the players in the lobby
*/
var lobbyData = [];


////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////	    FUNCTIONS   	////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

/**
 * Function returning the index of an element in a table
 * @param table, array to search in
 * @param key, index of the element to search
 * @param element, ellment to search
 * @returns int, index of the element in the table, -1 if not found
 */
function findIndex(table, key, element)											
{
	for (let i = 0; i < table.length; i++)
	{
		if (table[i][key] == element)
			return i;
	}
	return -1;
}

/**
 * Function printing the server data for debug purposes
 */
function printServerData()
{
	lobbyData.forEach((lobby, index) => 
	{
		console.log("CONTENU ACTUEL SUR LE SERVEUR : LOBBY", index + 1);

		console.log("------ Lobby ------");
		console.log("roomCode :", lobby["roomCode"]);
		console.log("nbPlayers :", lobby["nbPlayers"]);

		console.log("------ Players ------");
		lobby.players.forEach(player =>
		{
			console.log("login :", player["login"], ", isReady :", player["isReady"], ", socket :", player["socket"] == "" ? "bad" : "good");
		});
	});
}

/**
 * Function extracting the logins from a list of players linked with their sockets
 * @param players, array of players
 * @returns logins, array of logins
 */
function extractFromPlayers(players, key)
{
	let data = [];
	
	players.forEach(player => 
	{
		data.push(player[key]);
	});

	return data;
}

/**
 * Function removing a player from a lobby
 * @param socket, websocket of the player
 */
function removePlayer(socket)
{
	let indexLobby;
	let indexPlayerToRemove;
	for (indexLobby = 0; indexLobby < lobbyData.length; indexLobby++)
	{
		indexPlayerToRemove = findIndex(lobbyData[indexLobby].players, "socket", socket);
		if (indexPlayerToRemove != -1)
		{
			if (debugMode) console.log("success : lobby found, ready to quit")
			break;
		}
	}

	if (indexLobby >= lobbyData.length) 
	{
		if (debugMode) console.log("the player was not in a lobby, nothing changed");
		return;
	}

	lobbyData[indexLobby].players.splice(indexPlayerToRemove, 1);
	lobbyData[indexLobby].nbPlayers--;

	if (lobbyData[indexLobby].nbPlayers == 0)
	{
		lobbyData.splice(indexLobby, 1);

		if (debugMode) console.log("success : lobby deleted because everyone left")
		if (debugMode) printServerData();

		return;
	}

	dataJson["errorCode"] = "40";
	dataJson["errorMsg"] = "playerDisconnection";
	dataJson["playersInLobby"] = extractFromPlayers(lobbyData[indexLobby].players, "login");
	dataJson["statusInLobby"] = extractFromPlayers(lobbyData[indexLobby].players, "isReady");

	lobbyData[indexLobby].players.forEach((player, index) => 
	{
		dataJson["playerNumber"] = (index + 1).toString();
		player.socket.send(JSON.stringify(dataJson));

		if (debugMode) console.log("PAQUET ENVOYE AU JOUEUR", index + 1, "\n", dataJson);
	});
}

/*
 * Function that generate a 6 characters roomCode who does not already exists
 */
function generateRoomCode()
{
	let roomCodeValid = false;
	let roomCode = "";

	while (!roomCodeValid)
	{
		roomCode = "";

		for (let i = 0; i < 6; i++)
			roomCode += String.fromCharCode(65 + Math.random() * 26)

		if (findIndex(lobbyData, "roomCode", roomCode) == -1)
			roomCodeValid = true;
	}

	return roomCode;
}

////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////	WEBSOCKET CONNECTION	////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

wss.on('connection', function connection(client) 
{
	if (debugMode) console.log("\n\nnew connection !");  // sanity check

	////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////	 COMMUNICATION 	////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////

	client.on("message", (data) =>
	{
		// parse the data
		dataJson = JSON.parse(data);

		if (debugMode) console.log("\n\nPAQUET RECU :\n", dataJson);

		////////////////////////////////////////////////////////////////////////////////////////
		//////////////////////////////	    DATA CHEKING	////////////////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////

		/**
		 * Check the integrity of the data
		 * @var packetCorrect ,true if the packet is correct, false otherwise
		 * @var dataJson ,the data to check
		 */
		let packetCorrect = false;
		try
		{
			if (dataJson["login"] !== "" && 
			   (dataJson["inLobby"] === "true" || dataJson["inLobby"] === "false") &&
			   (dataJson["createLobby"] === "true" || dataJson["createLobby"] === "false") && 
			   (dataJson["isReady"] === "true" || dataJson["isReady"] === "false"))
				packetCorrect = true;
		}
		catch (error)  			// if the data is not correct, the server will crash, so we catch the error
		{}

		if (!packetCorrect)
		{
			dataJson["errorCode"] = "1";
			dataJson["errorMsg"] = "somethingWentWrong";
			client.send(JSON.stringify(dataJson));

			if (debugMode) console.log("error : somethingWentWrong");
			if (debugMode) console.log("PAQUET ENVOYE :\n", dataJson);
		
			return;
		}

		////////////////////////////////////////////////////////////////////////////////////////
		////////////////////////////	LOBBY CREATION	  //////////////////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////

		/**
		 * If creatLobby is true, create a new lobby in lobbyData, add the player to it
		 * and save the index of the lobby in lobbyIndex
		 * @var lobbyIndex ,index of the lobby in the lobbyData array
		 * @var dataJson ,working data
		 */

		// shouldn't be happening, but just in case
		if (dataJson["createLobby"] === "true" && dataJson["inLobby"] === "true")
		{
			dataJson["errorCode"] = "11";
			dataJson["errorMsg"] = "cannotCreateLobbyInLobby";
			client.send(JSON.stringify(dataJson));

			if (debugMode) console.log("error : cannotCreateLobbyInLobby");
			if (debugMode) console.log("PAQUET ENVOYE :\n", dataJson);

			return;
		}

		// we want to create a lobby
		if (dataJson["createLobby"] === "true")
		{
			let roomCode = generateRoomCode();
			let lobby = {roomCode : roomCode, nbPlayers : 0, players : []};
			lobbyData.push(lobby);

			dataJson["roomCode"] = roomCode;
			dataJson["createLobby"] = "false";

			if (debugMode) console.log("success : roomCreated, roomCode " + roomCode);
			if (debugMode) console.log("PAQUET ENVOYE :\n", dataJson);
		}

		////////////////////////////////////////////////////////////////////////////////////////
		////////////////////////////	LOBBY QUITTING	  //////////////////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////

		/**
		 * Handle lobby joining process
		 * if inLobby is false, lobbyIndex != -1 and there is less than 4 players in the lobby,
		 * add the player to the lobby and send the appropriate data
		 * @var lobbyIndex ,index of the lobby in the lobbyData array
		 * @var dataJson ,working data
		 * @send (error) 22 ,if the lobby doesn't exist
		 * @send dataJson ,if the lobby exists and the player is added to it
		 */
		
		// if we want to quit the lobby
		if (dataJson["inLobby"] === "false" && dataJson["roomCode"] == "")
		{
			dataJson["playerNumber"] = "";
			dataJson["errorCode"] = "24";
			dataJson["errorMsg"] = "lobbyQuitting";
			client.send(JSON.stringify(dataJson));
			
			if (debugMode) console.log("PAQUET ENVOYE :\n", dataJson);

			removePlayer(client);
			
			if (debugMode) console.log("success : lobby quitted successfully");
			if (debugMode) printServerData();
			
			return;
		}

		////////////////////////////////////////////////////////////////////////////////////////
		////////////////////////////	LOBBY JOINING	  //////////////////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////

		/**
		 * Handle lobby joining process
		 * if inLobby is false, lobbyIndex != -1 and there is less than 4 players in the lobby,
		 * add the player to the lobby and send the appropriate data
		 * @var lobbyIndex ,index of the lobby in the lobbyData array
		 * @var dataJson ,working data
		 * @send (error) 22 ,if the lobby doesn't exist
		 * @send dataJson ,if the lobby exists and the player is added to it
		 */

		// get the index of the lobby in the lobbyData array, if -1, the lobby doesn't exist or we want to quit the lobby
		let lobbyIndex = findIndex(lobbyData, "roomCode", dataJson["roomCode"]);

		// if we try to join a lobby that does not exists
		if (lobbyIndex == -1)
		{
			dataJson["errorCode"] = "22";
			dataJson["errorMsg"] = "lobbyNotFound";
			client.send(JSON.stringify(dataJson));

			if (debugMode) console.log("error : lobbyNotFound, roomCode " + dataJson["roomCode"]);
			if (debugMode) console.log("PAQUET ENVOYE :\n", dataJson);

			return;
		}
		
		// if we want to join the lobby that exists
		if (dataJson["inLobby"] === "false")
		{
			// but not if there is too many player
			if (lobbyData[lobbyIndex].nbPlayers >= 4)
			{
				dataJson["errorCode"] = "23";
				dataJson["errorMsg"] = "roomFull";
				client.send(JSON.stringify(dataJson));

				if (debugMode) console.log("error : roomFull, roomCode " + dataJson["roomCode"]);
				if (debugMode) console.log("PAQUET ENVOYE :\n", dataJson);

				return;
			}
			// we finally join the lobby
			else
			{
				let couple = {login : dataJson["login"], isReady : dataJson["isReady"], socket : client};

				lobbyData[lobbyIndex].nbPlayers++;
				lobbyData[lobbyIndex].players.push(couple);

				dataJson["playerNumber"] = lobbyData[lobbyIndex].nbPlayers.toString();
				dataJson["inLobby"] = "true";
				dataJson["playersInLobby"] = extractFromPlayers(lobbyData[lobbyIndex].players, "login");
				dataJson["statusInLobby"] = extractFromPlayers(lobbyData[lobbyIndex].players, "isReady");
				client.send(JSON.stringify(dataJson));

				if (debugMode) console.log("success : roomJoined, roomCode " + dataJson["roomCode"]);
				if (debugMode) console.log("PAQUET ENVOYE :\n", dataJson);

				dataJson["errorCode"] = "30";
				dataJson["errorMsg"] = "statusChanged";
			}
		}

		////////////////////////////////////////////////////////////////////////////////////////
		////////////////////////////	  DATA RELAY  	  //////////////////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////

		// checks for players who modified their status
		lobbyData[lobbyIndex].players.forEach((player, index, players) =>
		{
			if (player.socket === client && dataJson["isReady"] !== player.isReady)
			{
				players[index]["isReady"] = dataJson["isReady"];

				dataJson["errorCode"] = "30";
				dataJson["errorMsg"] = "statusChanged";
				dataJson["playersInLobby"] = extractFromPlayers(lobbyData[lobbyIndex].players, "login");
				dataJson["statusInLobby"] = extractFromPlayers(lobbyData[lobbyIndex].players, "isReady");
				client.send(JSON.stringify(dataJson));

				if (debugMode) console.log("warning : isReady changed for one player");
				if (debugMode) console.log("PAQUET ENVOYE :\n", dataJson);
			}
		});

		/**
		 * Send the data to every player in the lobby except the sender
		 * @var lobbyIndex ,index of the lobby in the lobbyData array
		 * @var lobbyData ,array containing all the lobbies
		 * @var dataJson ,working data
		 * @send dataJson ,to every player in the lobby except the sender
		 */
		lobbyData[lobbyIndex].players.forEach(player =>
		{
			if (player.socket !== client)
			{
				player.socket.send(JSON.stringify(dataJson));
			}
			else
			{
				if (debugMode) console.log("PAQUET ENVOYE AUX AUTRES JOUEURS DU LOBBY :\n", dataJson);
			}
		});
		
		if (debugMode) printServerData();
	});
	
	////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// 	DISCONNECTION 	////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////

	// dirty disconnection
	client.on("disconnect", () =>
	{
		if (debugMode) console.log("\n\ndirty disconnect detected !");
		if (debugMode) console.log("warning : playerDisconnection");
		
		removePlayer(client);

		if (debugMode) console.log("success : player deleted");
		if (debugMode) printServerData();
	});

	// clean disconnection
	client.on("close", () =>
	{
		if (debugMode) console.log("\n\nclean disconnect detected !");
		if (debugMode) console.log("warning : playerDisconnection");

		removePlayer(client);
		
		if (debugMode) console.log("success : player deleted");
		if (debugMode) printServerData();
	});
});

////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////	NOTHING TO SEE	////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
wss.on("listening", () =>
{
	console.log("Server is alive !");
});

httpServer.listen(10001, () => 
{
  	console.log("Server started on port 10001 !");
});